# -*- coding: utf-8 -*-
import json
import logging
import re

from django.contrib.auth.models import Permission
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils.safestring import mark_safe
from django.db import transaction
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.core.cache import cache

from mptt.models import MPTTModel, TreeForeignKey

from .managers import MenuItemManager
from .exceptions import InvalidJson, UnsupportedJsonData, MissingJsonRequiredProp

logger = logging.getLogger(__name__)


class MenuItem(MPTTModel):
    label = models.CharField(
        _("label"), max_length=50, help_text=_("Display name on the website.")
    )
    slug = models.SlugField(
        _("slug"), unique=True, help_text=_("Unique identifier for the menu voice.")
    )
    parent = TreeForeignKey(
        "self",
        verbose_name=_("parent"),
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="children",
    )
    link = models.CharField(_("link"), max_length=255, blank=True, null=True)
    title = models.CharField(_("title"), max_length=255, blank=True, null=True)
    order = models.IntegerField(_("order"))
    enabled = models.BooleanField(_("enabled"), default=True)
    login_required = models.BooleanField(
        _("login required"),
        default=False,
        help_text=_(
            "If this is checked, only logged-in users will be able to view the item."
        ),  # noqa
    )
    permissions = models.ManyToManyField(
        Permission,
        verbose_name=_("permissions"),
        blank=True,
        help_text=_(
            "If empty, the menu item will be publicly visible, otherwise only users with at least one of the selected permissions could see it."  # noqa
        ),
    )
    extras = models.CharField(
        _("extras"),
        max_length=255,
        blank=True,
        null=True,
        help_text=mark_safe(_("Comma separated list of extra-attributes, e.g.: <code>icon=\"fa fa-user\",data-tooltip=\"Go home!\"</code>")),
    )

    objects = MenuItemManager()

    active = False
    with_active = False

    def __str__(self):
        return "{}".format(self.label)

    class Meta:
        verbose_name = _("Menu item")
        verbose_name_plural = _("Menu items")

    class MPTTMeta:
        order_insertion_by = ("order",)

    @classmethod
    def from_json(cls, json_string):
        try:
            d = json.loads(json_string)
        except Exception as e:
            raise InvalidJson("Cannot parse provided json: %s" % str(e))

        if not isinstance(d, dict):
            raise UnsupportedJsonData(
                "Provided json should be a dictionary containing a single root voice"
            )

        with transaction.atomic():
            res = cls.create_item_from_dict(d)

        return res

    @classmethod
    def create_item_from_dict(cls, d, parent=None):
        if "label" not in d or "slug" not in d or "order" not in d:
            raise MissingJsonRequiredProp(
                "label, slug and order properties are mandatory"
            )

        item = MenuItem.objects.create(
            parent=parent,
            label=d.get("label"),
            slug=d.get("slug"),
            order=d.get("order"),
            link=d.get("link", None),
            title=d.get("title", None),
            enabled=d.get("enabled", True),
            login_required=d.get("login_required", False),
            extras=d.get("extras", None),
        )

        for permission_code in d.get("permissions", []):
            item.permissions.add(Permission.objects.get(codename=permission_code))

        for child in d.get("children", []):
            cls.create_item_from_dict(child, item)

        return item

    def is_active(self, path, prefix=""):
        return self.link and self.link == path

    def extras_dict(self):
        try:
            return dict(
                [re.sub('"|\'', '', i).strip() for i in s.split("=")]
                for s in self.extras.split(",")
            )
        except Exception:
            return {}

@receiver(post_save, sender=MenuItem)
@receiver(post_delete, sender=MenuItem)
def clear_lineup_cache(sender , **kwargs):
    cache.delete("lineup")



# import the logging library
import logging
import os

from django.core.management.base import BaseCommand, CommandError

from lineup.models import MenuItem

# Get an instance of a logger
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Imports a menu tree from a json file'

    def add_arguments(self, parser):
        parser.add_argument('json_path', type=str)

    def handle(self, *args, **options):
        path = options['json_path']
        if not os.path.isfile(path):
            raise CommandError('File %s does not exist' % path)

        # read file
        with open(path, 'r') as json_file:
            json_data = json_file.read()

        if json_data:
            try:
                MenuItem.from_json(json_data)
                self.stdout.write(self.style.SUCCESS('Successfully imported menu'))
            except Exception as e:
                raise CommandError('Cannot import menu: %s' % str(e))
        else:
            raise CommandError('File %s is empty' % path)



# Generated by Django 3.1.7 on 2021-04-25 13:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lineup', '0002_auto_20201213_0856'),
    ]

    operations = [
        migrations.AddField(
            model_name='menuitem',
            name='extras',
            field=models.CharField(blank=True, help_text="Add extra-paramaters on menu item eg: icon='fa- fa-user'", max_length=255, null=True, verbose_name='extras'),
        ),
    ]



from django import forms

class MenuItemForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        permissions = self.fields.get("permissions")
        if permissions:
            permissions.queryset = permissions.queryset.select_related(
                "content_type"
            )



# Generated by Django 3.1.4 on 2020-12-13 08:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lineup', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='menuitem',
            name='title',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='title'),
        ),
        migrations.AlterField(
            model_name='menuitem',
            name='enabled',
            field=models.BooleanField(default=True, verbose_name='enabled'),
        ),
        migrations.AlterField(
            model_name='menuitem',
            name='login_required',
            field=models.BooleanField(default=False, help_text='If this is checked, only logged-in users will be able to view the item.', verbose_name='login required'),
        ),
    ]



# Generated by Django 3.1.4 on 2020-12-11 09:14

from django.db import migrations, models
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):

    initial = True

    operations = [
        migrations.CreateModel(
            name='MenuItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(help_text='Display name on the website.', max_length=50, verbose_name='label')),
                ('slug', models.SlugField(help_text='Unique identifier for the menu voice.', unique=True, verbose_name='slug')),
                ('link', models.CharField(blank=True, max_length=255, null=True, verbose_name='link')),
                ('order', models.IntegerField(verbose_name='order')),
                ('enabled', models.BooleanField(verbose_name='enabled')),
                ('login_required', models.BooleanField(default=False, help_text='If this is checked, only logged-in users will be able to view the page.', verbose_name='login required')),
                ('lft', models.PositiveIntegerField(editable=False)),
                ('rght', models.PositiveIntegerField(editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(editable=False)),
                ('parent', mptt.fields.TreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='lineup.menuitem', verbose_name='parent')),
                ('permissions', models.ManyToManyField(blank=True, help_text='If empty, the menu item will be publicly visible, otherwise only users with at least one of the selected permissions could see it.', to='auth.Permission', verbose_name='permissions')),
            ],
            options={
                'verbose_name': 'Menu item',
                'verbose_name_plural': 'Menu items',
            },
        ),
    ]



# -*- coding: utf-8 -*-
import sys
import json
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from mptt.admin import MPTTModelAdmin

from .models import MenuItem
from .forms import MenuItemForm

baton = 'baton' in sys.modules

if baton:
    from baton.admin import RelatedDropdownFilter


class MenuItemInline(admin.StackedInline):
    '''
    Tabular Inline View for MenuItem
    '''
    model = MenuItem
    form = MenuItemForm
    extra = 1
    classes = ('collapse-entry', 'expand-first', )
    prepopulated_fields = {'slug': ('label',)}


@admin.register(MenuItem)
class MenuItemAdmin(MPTTModelAdmin):
    form = MenuItemForm
    change_list_template = "admin/lineup/menuitem/change_list.html"
    list_display = ('slug', 'label', 'parent', 'link', 'order', 'login_required', 'enabled', )
    list_filter = (('parent', RelatedDropdownFilter, ) if baton else 'parent', 'enabled', 'login_required', )
    list_editable = ('order', )
    search_fields = ('label', )
    filter_horizontal = ('permissions', )
    inlines = [MenuItemInline, ]
    prepopulated_fields = {'slug': ('label',)}
    fieldsets = (
        (_('Main'), {
            'fields': ('parent', 'label', 'slug', 'link', 'title', 'order', 'extras' ),
            'classes': ('baton-tabs-init', 'baton-tab-fs-permissions', 'baton-tab-inline-children', ),
            'description': _('A menu item without parent identifies a new menu.')

        }),
        (_('Visibility'), {
            'fields': ('enabled', 'login_required', 'permissions', ),
            'classes': ('tab-fs-permissions', ),
            'description': _('Yuo can decide to disable or restrict the visibility of this voice and consequently of all its children. Keep in mind that also if a child is publicly visible, but this voice requires a login, then the child will not be visible to not logged users. The same happens for permissions restrictions.')
        }),
    )

    def baton_cl_rows_attributes(self, request, cl):
        data = {}
        for item in MenuItem.objects.filter(parent__isnull=True):
            data[item.id] = {
                'class': 'table-info',
            }
        return json.dumps(data)



# -*- coding: utf-8 -*-
import logging
from django.core.cache import cache

from django import template
from django.contrib.auth.models import Permission
from django.db.models import Q, Count
from django.urls import LocalePrefixPattern

from ..models import MenuItem

register = template.Library()
logger = logging.getLogger(__name__)


def get_all_user_permissions_id_list(user):
    """
    Returns a flat list of all user permissions ids
    """
    # individual permissions
    perms = list(Permission.objects.filter(user=user).values_list("id", flat=True))
    # permissions that the user has via a group
    group_perms = list(
        Permission.objects.filter(group__user=user).values_list("id", flat=True)
    )
    return perms + group_perms


def remove_prefix(text, prefix): # TODO: introduce in python 3.9, so this can be removed
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


def set_active_voice(path, items):
    """
    Sets the active property to the active voice, and with_active to
    all its parents
    """

    l = LocalePrefixPattern(prefix_default_language=True)
    prefix = l.language_prefix
    for item in items:
        if item.get("instance").is_active(remove_prefix(path, prefix)):
            item["active"] = True
            parent = item.get("parent")
            while parent is not None:
                parent["has_active"] = True
                parent = parent.get("parent")


def create_tree(context, root, parent=None):
    # get current user
    user = context.get("user")

    # if root is not enabled, do not show children
    if not root.enabled:
        return {}

    # if root visibility restrictions and user is not authenticated do not show children
    if not user.is_authenticated and (
        root.login_required or root.permissions_count > 0
    ):  # noqa
        return {}

    if not user.is_authenticated:
        # unlogged user sees only public items
        items = root.children.enabled(login_required=False, permissions=None).annotate(permissions_count=Count("permissions"))
    elif user.is_superuser:
        # superuser sees all enabled items
        items = root.children.enabled().annotate(permissions_count=Count("permissions"))
    else:
        # logged in user which is not superuse should check for permissions
        permissions = context.get("lineup_permissions", None)
        if permissions is None:
            permissions = get_all_user_permissions_id_list(user)
            # put in context to avoid querying again at every iteration
            context["lineup_permissions"] = permissions

        items = root.children.enabled(
            Q(permissions__id__in=permissions) | Q(permissions=None)
        ).annotate(permissions_count=Count("permissions")).distinct()

    # parent needed to traverse upward for has-active functionality
    el = {"instance": root, "parent": parent}
    el["children"] = [create_tree(context, child, el) for child in items]

    # search active voice
    if "request" in context:
        path = context["request"].META["PATH_INFO"]
        set_active_voice(path, el["children"])

    return el


@register.inclusion_tag("lineup/menu.html", takes_context=True)
def lineup_menu(context, item):
    """
    Renders the whole tree if called recursively with different item param.
    - If item is a slug, the whole tree dictionary is created. and the first
    level items rendered.
    - If item is a dict, it just continues traversing rendering its children.
    """
    if isinstance(item, str):
        try:
            # cache depends on user (permissions) and path (active item)
            path = ""
            user = context.get("user")
            if "request" in context:
                path = context["request"].META["PATH_INFO"]
            t = cache.get_or_set("lineup", {}, None)
            key = "%s:%s:%s" % (item, user.id, path)
            if t.get(key, None) is None:
                root = MenuItem.objects.prefetch_related("children", "permissions").annotate(permissions_count=Count("permissions")).get(slug=item)
                tree = create_tree(context, root)
                items = tree.get("children", [])
                slug = item
                level = root.level
                t[key] = (items, slug, level)
                cache.set("lineup", t, None)
            else:
                (items, slug, level) = t[key]
        except MenuItem.DoesNotExist:
            logger.error("Provided lineup menu slug %s not found" % item)
            return context
    elif isinstance(item, dict):
        items = item.get("children")
        slug = item.get("instance").slug
        level = item.get("instance").level

    else:
        items = []
        slug = None
        level = None

    context["items"] = items
    context["slug"] = slug
    context["level"] = level

    return context


@register.inclusion_tag("lineup/breadcrumbs.html", takes_context=True)
def lineup_breadcrumbs(context, slug):
    """
    Renders menu breadcrumbs.
    """
    if "request" in context:
        path = context["request"].META["PATH_INFO"]
        active_items = MenuItem.objects.enabled(link=path)

        found = False
        items = []
        for active_item in active_items:
            items.append(active_item)
            parent = active_item.parent
            while parent is not None and parent.enabled:
                items.insert(0, parent)
                parent = parent.parent
            if items.pop(0).slug == slug:
                found = True
                break

        if found:
            context["items"] = items
        else:
            context["items"] = []
    return context



from django.urls import reverse
from django.views.generic import View
from django.shortcuts import redirect
from .models import MenuItem

class RebuildTreeView(View):
    def get(self, request):
        MenuItem.objects.rebuild()
        return redirect(reverse('admin:lineup_menuitem_changelist'))
